/************************************************************************ 
 * RSTP library - Rapid Spanning Tree (802.1t, 802.1w) 
 * Copyright (C) 2001-2003 Optical Access 
 * Author: Alex Rozin 
 * 
 * Portions copyright (C) 2004 Andy Myers, who modified this code to
 * be C++ and wrapped it in BridgeSim
 * 
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; version 2.1
 *
 * This code is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with RSTP library; see the file COPYING.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.
 **********************************************************************/
#include "global.h"
#include "Parser.h"
#include "Simulator.h"
#include "ArgParser.h"
#include "RapidSpanningTree.h"
#include "Log.h"

#include <iostream>
#include <istream>
#include <fstream>
#include <sstream>

void Parser::Parse(char *filename) {
  ifstream fstream(filename, ios_base::in);
  string line;

  if (!fstream.is_open()) {
    cerr << "Could not open file " << filename << endl;
    exit(1);
  }
  
  enum Section { SecBogus, SecPreamble, SecNodes, SecLinks, SecEvents };
  Section status = SecPreamble;
  unsigned lineNum = 0;

  /* read lines from the configuration file and process it */
  while(getline(fstream, line)) {
    lineNum++;
    line = trim(line);
    if (line.empty()) {
      continue;
    }

  if (line.find('[') != string::npos) // find configuration title
    {
      // it's a section header line
      if (line.find("nodes") != string::npos)	{
        cerr << "found a [nodes] section" << endl;
        status = SecNodes;
      } else if (line.find("links") != string::npos) {
        cerr << "found a [links] section" << endl;
        status = SecLinks;
      } else if (line.find("events") !=string::npos) {
        cerr << "found an [events] section" << endl;
        status = SecEvents;
      } else {
        cerr << "on line " << lineNum << " could not parse section header:"
             << endl << line << endl;
        exit(1);
      }
    } else {
      // it's a section body line
      switch (status) {
      case SecPreamble:
        ParsePreambleLine(line, lineNum);
        break;
      case SecNodes:
        ParseNodeLine(line, lineNum);
        break;
      case SecLinks:
        ParseLinkLine(line, lineNum);
        break;
      case SecEvents:
        ParseEventLine(line, lineNum);
        break;
      default:
        cerr << "Unrecognized section header on line " << lineNum << endl;
        exit(1);
      }
    }
  } // while getline

  fstream.close();
  cout<<"Parser done reading file " << filename << endl;
}

//  trim: get rid of extra blank both at the beginning and end of the line
string Parser::trim (string s)
{
  if (s.empty())
    return s;
  
  int pos_t = s.find('\r');
  if (pos_t!=-1)
    s.erase(pos_t,1);
  
  int pos_b = s.find_first_not_of(" ");
  if (pos_b!=0)
    s.erase(0, pos_b);
  
  unsigned int pos_e = s.find_last_not_of(" ");
  if (pos_e!=s.size()-1)
    s.erase(pos_e+1, s.length());
  return s;
}

void Parser::ParsePreambleLine(string& line, unsigned lineNum)
{
  istringstream str(line);

  string cmd;
  str >> cmd;

  if (cmd == "setdefault") {
    ArgParser ap;
    UnsignedArg helloTime("HelloTime", DEF_BR_HELLOT, ap);
    UnsignedArg maxAge("MaxAge", DEF_BR_MAXAGE, ap);
    UnsignedArg forwardDelay("ForwardDelay", DEF_BR_FWDELAY, ap);
    ap.Parse(str, lineNum);
    
    if (helloTime.val < MIN_BR_HELLOT) {
      cerr << "ERROR: HelloTime must be greater than or equal to " << MIN_BR_HELLOT << endl;
      exit(1);
    }
    if (helloTime.val > MAX_BR_HELLOT) {
      cerr << "ERROR: HelloTime must be less than or equal to " << MAX_BR_HELLOT << endl;
      exit(1);
    }
    if (maxAge.val < MIN_BR_MAXAGE) {
      cerr << "ERROR: MaxAge must be greater than or equal to " << MIN_BR_MAXAGE << endl;
      exit(1);
    }
    if (maxAge.val > MAX_BR_MAXAGE) {
      cerr << "ERROR: MaxAge must be less than or equal to " << MAX_BR_MAXAGE << endl;
      exit(1);
    }
    if (forwardDelay.val < MIN_BR_FWDELAY) {
      cerr << "ERROR: ForwardDelay must be greater than or equal to " << MIN_BR_FWDELAY 
           << endl;
      exit(1);
    }
    if (forwardDelay.val > MAX_BR_FWDELAY) {
      cerr << "ERROR: ForwardDelay must be less than or equal to " << MAX_BR_FWDELAY
           << endl;
      exit(1);
    }
    if (2 * (forwardDelay.val - 1) < maxAge.val) {
      cerr << "ERROR: 2*(ForwardDelay - 1) must be greater than or equal to MaxAge" << endl;
      exit(1);
    }
    if (maxAge.val < 2 * (helloTime.val + 1)) {
      cerr << "ERROR: MaxAge must be greater than or equal to 2*(HelloTime+1)" << endl;
      exit(1);
    }
    // XXXacm - move this into a function inside Simulator!
    DEF_BR_HELLOT = helloTime.val;
    DEF_BR_MAXAGE = maxAge.val;
    DEF_BR_FWDELAY = forwardDelay.val;
    stp_trace("reset bridge times: DEF_BR_HELLOT %d DEF_BR_MAXAGE %d DEF_BR_FWDELAY %d",
              DEF_BR_HELLOT, DEF_BR_MAXAGE, DEF_BR_FWDELAY);
  }
}

// line format: nodeID
// XXXacm - allow nodes to have arbitrary strings for names
void Parser::ParseNodeLine(string& line, unsigned lineNum)
{
  istringstream str(line);
  unsigned nodeid = 0;

  str >> nodeid;
  if (nodeid == 0) {
    cerr << "could not parse node id in line " << lineNum << " of input:"
         << endl << line << endl;
    exit(1);
  }

  ArgParser ap;
  DoubleArg offset("offset", 0, ap);

  ap.Parse(str, lineNum);

  if (offset.val < 0 || offset.val >= 1.0) {
    cerr << "offset must be greater than or equal to 0 or less than 1.0 on input line "
         << lineNum << endl;
    exit(1);
  }

  simulator->AddNode(nodeid, offset.val);  
}

// line format: nodeid1 nodeid2 delayInSeconds
void Parser::ParseLinkLine(string& line, unsigned lineNum)
{
  istringstream str(line);
  unsigned nodeA = 0, nodeB = 0;

  str >> nodeA >> nodeB;

  ArgParser ap;
  DoubleArg delay("delay", -1.0, ap);
  ap.Parse(str, lineNum);

  if (nodeA == 0 || nodeB == 0 || delay.val < 0) {
    cerr << "could not parse link info on line " << lineNum << " of input:"
         << endl << line << endl;
    exit(1);
  }
  simulator->AddLink(nodeA, nodeB, delay.val);
}

// line format: timeInSeconds EventType [type-specitic args]
void Parser::ParseEventLine(string& line, unsigned lineNum)
{
  istringstream str(line);
  double when = -1;
  string evType = "bogus";

  str >> when >> evType;

  if (when == -1 || evType == "bogus") {
    cerr << "could not parse event on line " << lineNum << " of input:"
         << endl << line << endl;
    exit(1);
  }

  if (evType == "linkdown") {
    ArgParser ap;
    UnsignedArg nodeA("nodeA", 0, ap);
    UnsignedArg nodeB("nodeB", 0, ap);
    DoubleArg delayA("delayA", 0, ap);
    DoubleArg delayB("delayB", 0, ap);
    ap.Parse(str, lineNum);

    if (delayA.val < 0 || delayB.val < 0) {
      cerr << "on line " << lineNum << " delayA and delayB must be non-negative"
           << endl;
      exit(1);
    }

    simulator->AddLinkDownEvent(when, nodeA.val, nodeB.val, delayA.val, delayB.val);
  } else if (evType == "simend") {
    simulator->AddSimulationEndEvent(when);
  } else {
    cerr << "could not parse event type " << evType << " on line " << lineNum 
         << " of input:" << endl << line << endl;
    exit(1);
  }
}

